"""
Provides a simple implementation of timers run inside asyncio event loops.
"""

import asyncio
import contextlib
import enum
import functools
from typing import Callable, Optional
from unittest import mock

from .compat import get_running_loop
from .taskgroup import TaskGroup

__all__ = (
    'create_timer',
    'TimerDelayPolicy',
    'VirtualClock',
)


class TimerDelayPolicy(enum.Enum):
    """
    An enumeration of supported policies for when the timer function takes
    longer on each tick than the given timer interval.
    """
    DEFAULT = 0
    CANCEL = 1


def create_timer(cb: Callable[[float], None], interval: float,
                 delay_policy: TimerDelayPolicy = TimerDelayPolicy.DEFAULT,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> asyncio.Task:
    """
    Schedule a timer with the given callable and the interval in seconds.
    The interval value is also passed to the callable.
    If the callable takes longer than the timer interval, all accumulated
    callable's tasks will be cancelled when the timer is cancelled.

    Args:
        cb: TODO - fill argument descriptions

    Returns:
        You can stop the timer by cancelling the returned task.
    """
    if loop is None:
        loop = get_running_loop()

    async def _timer():
        fired_tasks = []
        try:
            async with TaskGroup() as task_group:
                while True:
                    if delay_policy == TimerDelayPolicy.CANCEL:
                        for t in fired_tasks:
                            if not t.done():
                                t.cancel()
                        await asyncio.gather(*fired_tasks, return_exceptions=True)
                        fired_tasks.clear()
                    else:
                        fired_tasks[:] = [t for t in fired_tasks if not t.done()]
                    t = task_group.create_task(cb(interval=interval))
                    fired_tasks.append(t)
                    await asyncio.sleep(interval)
        except asyncio.CancelledError:
            pass
        finally:
            await asyncio.sleep(0)

    return loop.create_task(_timer())


class VirtualClock:
    """
    Provide a virtual clock for an asyncio event loop
    which makes timing-based tests deterministic and instantly completed.
    """

    def __init__(self) -> None:
        self.vtime = 0.0

    def virtual_time(self) -> float:
        """
        Return the current virtual time.
        """
        return self.vtime

    def _virtual_select(self, orig_select, timeout):
        self.vtime += timeout
        return orig_select(0)  # override the timeout to zero

    @contextlib.contextmanager
    def patch_loop(self):
        """
        Override some methods of the current event loop
        so that sleep instantly returns while proceeding the virtual clock.
        """
        loop = get_running_loop()
        with mock.patch.object(
            loop._selector,
            'select',
            new=functools.partial(self._virtual_select, loop._selector.select),
        ), \
            mock.patch.object(
            loop,
            'time',
            new=self.virtual_time,
        ):
            yield



# Import submodules only when installed properly
from . import (
    context,
    defer as _defer,
    func,
    iter as _iter,
    timer,
    taskgroup,
    server,
)

import pkgutil
_version_data = pkgutil.get_data("aiotools", "VERSION")
if _version_data is None:
    __version__ = '0.0.dev'
else:
    __version__ = _version_data.decode('utf8').strip()

__all__ = (
    *context.__all__,
    *_defer.__all__,
    *func.__all__,
    *_iter.__all__,
    *server.__all__,
    *taskgroup.__all__,
    *timer.__all__,
    '__version__',
)

from .context import *  # noqa
from .defer import *    # noqa
from .func import *     # noqa
from .iter import *     # noqa
from .timer import *    # noqa
from .server import *   # noqa



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# This file is included in the aiotools open source project
# given that EdgeDB's Apache 2.0 license is compatible with
# aiotools' MIT license.
# The original source code is taken from:
# https://github.com/edgedb/edgedb-python/blob/bcbe005/edgedb/_taskgroup.py


import asyncio
import functools
import itertools
import textwrap
import traceback


__all__ = (
    'MultiError',
    'TaskGroup',
    'TaskGroupError',
)


class TaskGroup:
    """
    Provides a guard against a group of tasks spawend via its :meth:`create_task`
    method instead of the vanilla fire-and-forgetting :meth:`asyncio.create_task`.

    See the motivation and rationale in `the trio's documentation
    <https://trio.readthedocs.io/en/stable/reference-core.html#nurseries-and-spawning>`_.
    """

    def __init__(self, *, name=None):
        if name is None:
            self._name = f'tg-{_name_counter()}'
        else:
            self._name = str(name)

        self._entered = False
        self._exiting = False
        self._aborting = False
        self._loop = None
        self._parent_task = None
        self._parent_cancel_requested = False
        self._tasks = set()
        self._unfinished_tasks = 0
        self._errors = []
        self._base_error = None
        self._on_completed_fut = None

    def get_name(self):
        return self._name

    def __repr__(self):
        msg = f'<TaskGroup {self._name!r}'
        if self._tasks:
            msg += f' tasks:{len(self._tasks)}'
        if self._unfinished_tasks:
            msg += f' unfinished:{self._unfinished_tasks}'
        if self._errors:
            msg += f' errors:{len(self._errors)}'
        if self._aborting:
            msg += ' cancelling'
        elif self._entered:
            msg += ' entered'
        msg += '>'
        return msg

    async def __aenter__(self):
        if self._entered:
            raise RuntimeError(
                f"TaskGroup {self!r} has been already entered")
        self._entered = True

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if hasattr(asyncio, 'current_task'):
            self._parent_task = asyncio.current_task(self._loop)
        else:
            self._parent_task = asyncio.Task.current_task(self._loop)

        if self._parent_task is None:
            raise RuntimeError(
                f'TaskGroup {self!r} cannot determine the parent task')
        self._patch_task(self._parent_task)

        return self

    async def __aexit__(self, et, exc, tb):
        self._exiting = True
        propagate_cancelation = False

        if (exc is not None and
                self._is_base_error(exc) and
                self._base_error is None):
            self._base_error = exc

        if et is asyncio.CancelledError:
            if self._parent_cancel_requested:
                # Only if we did request task to cancel ourselves
                # we mark it as no longer cancelled.
                self._parent_task.__cancel_requested__ = False
            else:
                propagate_cancelation = True

        if et is not None and not self._aborting:
            # Our parent task is being cancelled:
            #
            #    async with TaskGroup() as g:
            #        g.create_task(...)
            #        await ...  # <- CancelledError
            #
            if et is asyncio.CancelledError:
                propagate_cancelation = True

            # or there's an exception in "async with":
            #
            #    async with TaskGroup() as g:
            #        g.create_task(...)
            #        1 / 0
            #
            self._abort()

        # We use while-loop here because "self._on_completed_fut"
        # can be cancelled multiple times if our parent task
        # is being cancelled repeatedly (or even once, when
        # our own cancellation is already in progress)
        while self._unfinished_tasks:
            if self._on_completed_fut is None:
                self._on_completed_fut = self._loop.create_future()

            try:
                await self._on_completed_fut
            except asyncio.CancelledError:
                if not self._aborting:
                    # Our parent task is being cancelled:
                    #
                    #    async def wrapper():
                    #        async with TaskGroup() as g:
                    #            g.create_task(foo)
                    #
                    # "wrapper" is being cancelled while "foo" is
                    # still running.
                    propagate_cancelation = True
                    self._abort()

            self._on_completed_fut = None

        assert self._unfinished_tasks == 0
        self._on_completed_fut = None  # no longer needed

        if self._base_error is not None:
            raise self._base_error

        if propagate_cancelation:
            # The wrapping task was cancelled; since we're done with
            # closing all child tasks, just propagate the cancellation
            # request now.
            raise asyncio.CancelledError()

        if et is not None and et is not asyncio.CancelledError:
            self._errors.append(exc)

        if self._errors:
            # Exceptions are heavy objects that can have object
            # cycles (bad for GC); let's not keep a reference to
            # a bunch of them.
            errors = self._errors
            self._errors = None

            me = TaskGroupError('unhandled errors in a TaskGroup',
                                errors=errors)
            raise me from None

    def create_task(self, coro):
        if not self._entered:
            raise RuntimeError(f"TaskGroup {self!r} has not been entered")
        if self._exiting:
            raise RuntimeError(f"TaskGroup {self!r} is awaiting in exit")
        task = self._loop.create_task(coro)
        task.add_done_callback(self._on_task_done)
        self._unfinished_tasks += 1
        self._tasks.add(task)
        return task

    def _is_base_error(self, exc):
        assert isinstance(exc, BaseException)
        return not isinstance(exc, Exception)

    def _patch_task(self, task):
        # In Python 3.8 we'll need proper API on asyncio.Task to
        # make TaskGroups possible. We need to be able to access
        # information about task cancellation, more specifically,
        # we need a flag to say if a task was cancelled or not.
        # We also need to be able to flip that flag.

        def _task_cancel(task, orig_cancel):
            task.__cancel_requested__ = True
            return orig_cancel()

        if hasattr(task, '__cancel_requested__'):
            return

        task.__cancel_requested__ = False
        # confirm that we were successful at adding the new attribute:
        assert not task.__cancel_requested__

        orig_cancel = task.cancel
        task.cancel = functools.partial(_task_cancel, task, orig_cancel)

    def _abort(self):
        self._aborting = True

        for t in self._tasks:
            if not t.done():
                t.cancel()

    def _on_task_done(self, task):
        self._unfinished_tasks -= 1
        assert self._unfinished_tasks >= 0

        if self._exiting and not self._unfinished_tasks:
            if not self._on_completed_fut.done():
                self._on_completed_fut.set_result(True)

        if task.cancelled():
            return

        exc = task.exception()
        if exc is None:
            return

        self._errors.append(exc)
        if self._is_base_error(exc) and self._base_error is None:
            self._base_error = exc

        if self._parent_task.done():
            # Not sure if this case is possible, but we want to handle
            # it anyways.
            self._loop.call_exception_handler({
                'message': f'Task {task!r} has errored out but its parent '
                           f'task {self._parent_task} is already completed',
                'exception': exc,
                'task': task,
            })
            return

        self._abort()
        if not self._parent_task.__cancel_requested__:
            # If parent task *is not* being cancelled, it means that we want
            # to manually cancel it to abort whatever is being run right now
            # in the TaskGroup.  But we want to mark parent task as
            # "not cancelled" later in __aexit__.  Example situation that
            # we need to handle:
            #
            #    async def foo():
            #        try:
            #            async with TaskGroup() as g:
            #                g.create_task(crash_soon())
            #                await something  # <- this needs to be canceled
            #                                 #    by the TaskGroup, e.g.
            #                                 #    foo() needs to be cancelled
            #        except Exception:
            #            # Ignore any exceptions raised in the TaskGroup
            #            pass
            #        await something_else     # this line has to be called
            #                                 # after TaskGroup is finished.
            self._parent_cancel_requested = True
            self._parent_task.cancel()


class MultiError(Exception):
    """
    Represents a collection of errors raised inside a task group.
    Callers may iterate over the errors using the ``__erros__`` attribute.
    """

    def __init__(self, msg, *args, errors=()):
        if errors:
            types = set(type(e).__name__ for e in errors)
            msg = f'{msg}; {len(errors)} sub errors: ({", ".join(types)})'
            for er in errors:
                msg += f'\n + {type(er).__name__}: {er}'
                if er.__traceback__:
                    er_tb = ''.join(traceback.format_tb(er.__traceback__))
                    er_tb = textwrap.indent(er_tb, ' | ')
                    msg += f'\n{er_tb}\n'
        super().__init__(msg, *args)
        self.__errors__ = tuple(errors)

    def get_error_types(self):
        return {type(e) for e in self.__errors__}

    def __reduce__(self):
        return (type(self), (self.args,), {'__errors__': self.__errors__})


class TaskGroupError(MultiError):
    """
    An alias to :exc:`MultiError`.
    """
    pass


_name_counter = itertools.count(1).__next__



"""
Based on :doc:`aiotools.context`, this module provides an automated lifecycle
management for multi-process servers with explicit initialization steps and
graceful shutdown steps.

.. function:: server(func)

    A decorator wrapper for :class:`AsyncServerContextManager`.

    Usage example:

    .. code:: python

       @aiotools.server
       async def myserver(loop, pidx, args):
           await do_init(args)
           stop_sig = yield
           if stop_sig == signal.SIGINT:
               await do_graceful_shutdown()
           else:
               await do_forced_shutdown()

       aiotools.start_server(myserver, ...)
"""

import asyncio
from contextlib import (
    AbstractContextManager, ContextDecorator,
)
import functools
import inspect
import logging
import multiprocessing as mp, multiprocessing.synchronize
import threading
import os
import signal
import struct
import sys
import time
from typing import (
    Any, Optional, Union,
    Callable, Iterable,
    Tuple, Sequence,
    Set,
)

from .context import AbstractAsyncContextManager

if sys.version_info < (3, 8, 0):
    from typing_extensions import Literal
else:
    from typing import Literal  # type: ignore

__all__ = (
    'main',
    'start_server',
    'AsyncServerContextManager',
    'ServerMainContextManager',
    'InterruptedBySignal',
)

log = logging.getLogger(__name__)

# for threaded mode
_children_ctxs = []
_children_loops = []
_children_ffs = []
_children_lock = threading.Lock()


class InterruptedBySignal(BaseException):
    """
    A new :class:`BaseException` that represents interruption by an arbitrary UNIX
    signal.

    Since this is a :class:`BaseException` instead of :class:`Exception`, it behaves
    like :class:`KeyboardInterrupt` and :class:`SystemExit` exceptions (i.e.,
    bypassing except clauses catching the :class:`Exception` type only)

    The first argument of this exception is the signal number received.
    """
    pass


class AsyncServerContextManager(AbstractAsyncContextManager):
    """
    A modified version of :func:`contextlib.asynccontextmanager`.

    The implementation detail is mostly taken from the ``contextlib`` standard
    library, with a minor change to inject ``self.yield_return`` into the wrapped
    async generator.
    """

    yield_return: Optional[signal.Signals]

    def __init__(self, func: Callable[..., Any], args, kwargs):
        if not inspect.isasyncgenfunction(func):
            raise RuntimeError('Context manager function must be '
                               'an async-generator')
        self._agen = func(*args, **kwargs)
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.yield_return = None

    async def __aenter__(self):
        try:
            return (await self._agen.__anext__())
        except StopAsyncIteration:
            raise RuntimeError("async-generator didn't yield") from None

    async def __aexit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            try:
                # Here is the modified part.
                await self._agen.asend(self.yield_return)
            except StopAsyncIteration:
                return
            else:
                raise RuntimeError("async-generator didn't stop") from None
        else:
            if exc_value is None:
                exc_value = exc_type()
            try:
                await self._agen.athrow(exc_type, exc_value, tb)
                raise RuntimeError("async-generator didn't stop after athrow()")
            except StopAsyncIteration as exc_new_value:
                return exc_new_value is not exc_value
            except RuntimeError as exc_new_value:
                if exc_new_value is exc_value:
                    return False
                if isinstance(exc_value, (StopIteration, StopAsyncIteration)):
                    if exc_new_value.__cause__ is exc_value:
                        return False
                raise
            except (BaseException, asyncio.CancelledError) as exc:
                if exc is not exc_value:
                    raise


class ServerMainContextManager(AbstractContextManager,
                               ContextDecorator):
    """
    A modified version of :func:`contextlib.contextmanager`.

    The implementation detail is mostly taken from the ``contextlib`` standard
    library, with a minor change to inject ``self.yield_return`` into the wrapped
    generator.
    """

    yield_return: Optional[signal.Signals]

    def __init__(self, func, args, kwargs):
        self.gen = func(*args, **kwargs)
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.yield_return = None

    def __enter__(self):
        del self.args, self.kwargs, self.func
        try:
            return next(self.gen)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __exit__(self, type, value, traceback):
        if type is None:
            try:
                self.gen.send(self.yield_return)
            except StopIteration:
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                value = type()
            try:
                self.gen.throw(type, value, traceback)
            except StopIteration as exc:
                return exc is not value
            except RuntimeError as exc:
                if exc is value:
                    return False
                if type is StopIteration and exc.__cause__ is value:
                    return False
                raise
            except:  # noqa
                if sys.exc_info()[1] is value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")


# This is a dirty hack to implement "module callable".
# NOTE: only works in Python 3.5 or higher.

def _server_ctxmgr(func):
    @functools.wraps(func)
    def helper(*args, **kwargs):
        return AsyncServerContextManager(func, args, kwargs)
    return helper


class _ServerModule(sys.modules[__name__].__class__):  # type: ignore
    def __call__(self, func):
        return _server_ctxmgr(func)


sys.modules[__name__].__class__ = _ServerModule


def _main_ctxmgr(func):
    """
    A decorator wrapper for :class:`ServerMainContextManager`

    Usage example:

    .. code:: python

       @aiotools.main
       def mymain():
           server_args = do_init()
           stop_sig = yield server_args
           if stop_sig == signal.SIGINT:
               do_graceful_shutdown()
           else:
               do_forced_shutdown()

       aiotools.start_server(..., main_ctxmgr=mymain, ...)
    """
    @functools.wraps(func)
    def helper(*args, **kwargs):
        return ServerMainContextManager(func, args, kwargs)
    return helper


main = _main_ctxmgr


def _worker_main(
        worker_actxmgr: Callable[
            [asyncio.AbstractEventLoop, int, Sequence[Any]],
            AsyncServerContextManager],
        threaded: bool,
        stop_signals: Set[signal.Signals],
        intr_pipe: Union[threading.Event, mp.synchronize.Event],
        proc_idx: int,
        args: Sequence[Any]):

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    interrupted = asyncio.Event()
    ctx = worker_actxmgr(loop, proc_idx, args)
    forever_future = loop.create_future()

    def handle_stop_signal(signum):
        if interrupted.is_set():
            pass
        else:
            interrupted.set()
            ctx.yield_return = signum
            forever_future.cancel()

    if not threaded:
        for signum in stop_signals:
            loop.add_signal_handler(
                signum,
                functools.partial(handle_stop_signal, signum))
        # Allow the worker to be interrupted during initialization
        # (in case of initialization failures in other workers)
        signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)
    else:
        with _children_lock:
            _children_ctxs.append(ctx)
            _children_loops.append(loop)
            _children_ffs.append(forever_future)

    async def _wrapped_worker():
        err_ctx = 'enter'
        try:
            async with ctx:
                err_ctx = 'body'
                try:
                    await forever_future
                except asyncio.CancelledError:
                    pass
                finally:
                    err_ctx = 'exit'
        except Exception:
            if err_ctx != 'body':
                err_ctx_str = 'initialization' if err_ctx == 'enter' else 'shutdown'
                log.exception(f'Worker {proc_idx}: '
                              f'Error during context manager {err_ctx_str}')
                wfd = intr_pipe if threaded else intr_pipe.fileno()
                os.write(wfd, struct.pack('i', proc_idx))
                # FIXME: this sleep is required to prevent hang-up in Linux
                time.sleep(0.2)
            raise

    try:
        loop.run_until_complete(_wrapped_worker())
    finally:
        try:
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            loop.close()


def _extra_main(extra_func, threaded, stop_signals, intr_event, proc_idx, args):
    interrupted: Union[threading.Event, mp.synchronize.Event]
    if threaded:
        interrupted = threading.Event()
    else:
        interrupted = mp.Event()
    if not threaded:

        # Since signals only work for the main thread in Python,
        # extra processes in use_threading=True mode should check
        # the intr_event by themselves (probably in their loops).

        def raise_stop(signum, frame):
            if interrupted.is_set():
                pass
            else:
                interrupted.set()
                if signum == signal.SIGINT:
                    raise KeyboardInterrupt
                elif signum == signal.SIGTERM:
                    raise SystemExit
                else:
                    raise InterruptedBySignal(signum)

        # restore signal handler.
        for signum in stop_signals:
            signal.signal(signum, raise_stop)
        signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)
        intr_event = None

    try:
        if not interrupted.is_set():
            extra_func(intr_event, proc_idx, args)
    except (SystemExit, KeyboardInterrupt, InterruptedBySignal):
        log.warning(f'extra_proc[{proc_idx}] did not handle stop signals.')
    finally:
        if not threaded:
            # same as in _worker_main()
            signal.pthread_sigmask(signal.SIG_BLOCK, stop_signals)


def start_server(
        worker_actxmgr: Callable[
            [asyncio.AbstractEventLoop, int, Sequence[Any]],
            AsyncServerContextManager],
        main_ctxmgr: Optional[Callable[[], ServerMainContextManager]] = None,
        extra_procs: Iterable[Callable] = tuple(),
        stop_signals: Iterable[signal.Signals] = (
            signal.SIGINT,
            signal.SIGTERM),
        num_workers: int = 1,
        use_threading: bool = False,
        start_method: Literal['spawn', 'fork', 'forkserver'] = None,
        args: Iterable[Any] = tuple()):
    """
    Starts a multi-process server where each process has their own individual
    asyncio event loop.  Their lifecycles are automantically managed -- if the
    main program receives one of the signals specified in ``stop_signals`` it
    will initiate the shutdown routines on each worker that stops the event
    loop gracefully.

    Args:
        worker_actxmgr: An asynchronous context manager that dicates the
                        initialization and shutdown steps of each worker.
                        It should accept the following three arguments:

                        * **loop**: the asyncio event loop created and set
                          by aiotools
                        * **pidx**: the 0-based index of the worker
                          (use this for per-worker logging)
                        * **args**: a concatenated tuple of values yielded by
                          **main_ctxmgr** and the user-defined arguments in
                          **args**.

                        aiotools automatically installs an interruption handler
                        that calls ``loop.stop()`` to the given event loop,
                        regardless of using either threading or
                        multiprocessing.

        main_ctxmgr: An optional context manager that performs global
                     initialization and shutdown steps of the whole program.
                     It may yield one or more values to be passed to worker
                     processes along with **args** passed to this function.
                     There is no arguments passed to those functions since
                     you can directly access ``sys.argv`` to parse command
                     line arguments and/or read user configurations.

        extra_procs: An iterable of functions that consist of extra processes
                     whose lifecycles are synchronized with other workers.

                     You should write the shutdown steps of them differently
                     depending on the value of **use_threading** argument.

                     If it is ``False`` (default), they will get
                     a :class:`BaseException` depending on the received stop signal
                     number, either :class:`KeyboardInterrupt` (for SIGINT),
                     :class:`SystemExit` (for SIGTERM), or
                     :class:`InterruptedBySignal` (otherwise).

                     If it is ``True``, they should check their **intr_event**
                     argument periodically because there is no way to install
                     signal handlers in Python threads (only the main thread
                     can install signal handlers).

                     It should accept the following three arguments:

                     * **intr_event**: :class:`threading.Event` object that
                       signals the interruption of the main thread (only
                       available when **use_threading** is ``True``; otherwise
                       it is set to ``None``)
                     * **pidx**: same to **worker_actxmgr** argument
                     * **args**: same to **worker_actxmgr** argument

        stop_signals: A list of UNIX signals that the main program to
                      recognize as termination signals.

        num_workers: The number of children workers.

        use_threading: Use :mod:`threading` instead of :mod:`multiprocessing`.
                       In this case, the GIL may become the performance
                       bottleneck.  Set this ``True`` only when you know what
                       you are going to do.  Note that this changes the way
                       to write user-defined functions passed as **extra_procs**.

        start_method: Change the start method when :mod:`multiprocessing` is used.
                      The default is same to what :mod:`multiprocessing` uses.
                      Only effective when **use_threading** is ``False``.

                      Note that if there are other libraries that rely on
                      :mod:`multiprocessing` called before aiotools, you need to
                      invoke :func:`multiprocessing.set_start_method()` earlier than
                      both aiotools and such libraries *without* setting this
                      argument as it changes the global context of
                      :mod:`multiprocessing`.

        args: The user-defined arguments passed to workers and extra
              processes.  If **main_ctxmgr** yields one or more values,
              they are *prepended* to this user arguments when passed to
              workers and extra processes.

    Returns:
        None

    .. versionchanged:: 0.3.2

       The name of argument **num_proc** is changed to **num_workers**.
       Even if **num_workers** is 1, a child is created instead of
       doing everything at the main thread.

    .. versionadded:: 0.3.2

       The argument ``extra_procs`` and ``main_ctxmgr``.

    .. versionadded:: 0.4.0

       Now supports use of threading instead of multiprocessing via
       **use_threading** option.

    .. versionchanged:: 0.8.0

       Now **worker_actxmgr** must be an instance of
       :class:`AsyncServerContextManager` or async generators decorated by
       ``@aiotools.server``.

       Now **main_ctxmgr** must be an instance of :class:`ServerMainContextManager`
       or plain generators decorated by ``@aiotools.main``.

       The usage is same to asynchronous context managers, but optionally you can
       distinguish the received stop signal by retrieving the return value of the
       ``yield`` statement.

       In **extra_procs** in non-threaded mode, stop signals are converted into
       either one of :class:`KeyboardInterrupt`, :class:`SystemExit`, or
       :class:`InterruptedBySignal` exception.

    .. versionadded:: 0.8.4

       **start_method** argument can be set to change the subprocess spawning
       implementation.
    """

    @_main_ctxmgr
    def noop_main_ctxmgr():
        yield

    def create_child(*args, **kwargs):
        if use_threading:
            return threading.Thread(*args, **kwargs)
        else:
            return mp.Process(*args, **kwargs)

    assert stop_signals

    if hasattr(asyncio, 'get_running_loop'):
        # only for Python 3.7+
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            # there should be none.
            pass
        else:
            raise RuntimeError(
                'aiotools.start_server() cannot be called inside '
                'a running event loop.')

    if main_ctxmgr is None:
        main_ctxmgr = noop_main_ctxmgr
    if not use_threading and start_method is not None:
        mp.set_start_method(start_method)

    children = []
    _children_ctxs.clear()
    _children_loops.clear()
    _children_ffs.clear()
    intr_event: Union[threading.Event, mp.synchronize.Event]
    if use_threading:
        intr_event = threading.Event()
    else:
        intr_event = mp.Event()
    sigblock_mask = frozenset(stop_signals)

    main_ctx = main_ctxmgr()

    # temporarily block signals and register signal handlers to mainloop
    signal.pthread_sigmask(signal.SIG_BLOCK, sigblock_mask)

    mainloop = asyncio.new_event_loop()
    asyncio.set_event_loop(mainloop)

    # to make subprocess working in child threads
    try:
        asyncio.get_child_watcher()
    except NotImplementedError:
        pass  # for uvloop

    # build a main-to-worker interrupt channel using signals
    def handle_stop_signal(signum: signal.Signals) -> None:
        main_ctx.yield_return = signum  # type: ignore
        if use_threading:
            with _children_lock:
                for c in _children_ctxs:
                    c.yield_return = signum
                for child_loop, ff in zip(_children_loops, _children_ffs):
                    child_loop.call_soon_threadsafe(ff.cancel)
            intr_event.set()
        else:
            os.killpg(0, signum)
        mainloop.stop()

    for signum in stop_signals:
        mainloop.add_signal_handler(
            signum,
            functools.partial(handle_stop_signal, signum))

    # build a reliable worker-to-main interrupt channel using a pipe
    # (workers have no idea whether the main interrupt is enabled/disabled)
    def handle_child_interrupt(fd: int) -> None:
        child_idx = struct.unpack('i', os.read(fd, 4))[0]  # noqa
        log.debug(f'Child {child_idx} has interrupted the main program.')
        # self-interrupt to initiate the main-to-worker interrupts
        signal.pthread_sigmask(signal.SIG_UNBLOCK, {signal.SIGINT})
        os.kill(0, signal.SIGINT)

    child_intr_pipe: Union[Tuple[int, int],
                           Tuple[mp.connection.Connection, mp.connection.Connection]]
    if use_threading:
        child_intr_pipe = os.pipe()
        rfd = child_intr_pipe[0]
    else:
        child_intr_pipe = mp.Pipe()
        rfd = child_intr_pipe[0].fileno()
    mainloop.add_reader(rfd, handle_child_interrupt, rfd)

    # start
    with main_ctx as main_args:

        # retrieve args generated by the user-defined main
        if main_args is None:
            main_args = tuple()
        if not isinstance(main_args, tuple):
            main_args = (main_args, )

        # spawn managed async workers
        for i in range(num_workers):
            p = create_child(target=_worker_main, daemon=True,
                             args=(worker_actxmgr, use_threading, stop_signals,
                                   child_intr_pipe[1], i,
                                   main_args + args))
            p.start()
            children.append(p)

        # spawn extra workers
        for i, f in enumerate(extra_procs):
            p = create_child(target=_extra_main, daemon=True,
                             args=(f, use_threading, stop_signals,
                                   intr_event, num_workers + i,
                                   main_args + args))
            p.start()
            children.append(p)
        try:
            # unblock the stop signals for user/external interrupts.
            signal.pthread_sigmask(signal.SIG_UNBLOCK, sigblock_mask)

            # run!
            mainloop.run_forever()

            # if interrupted, wait for workers to finish.
            for child in children:
                child.join()
        finally:
            try:
                mainloop.run_until_complete(mainloop.shutdown_asyncgens())
            finally:
                mainloop.close()
                asyncio.set_event_loop(None)



import asyncio


if hasattr(asyncio, 'get_running_loop'):
    get_running_loop = asyncio.get_running_loop
else:
    get_running_loop = asyncio.get_event_loop


if hasattr(asyncio, 'all_tasks'):
    all_tasks = asyncio.all_tasks
else:
    all_tasks = asyncio.Task.all_tasks



__all__ = (
    'aiter',
)


_sentinel = object()


async def aiter(obj, sentinel=_sentinel):
    """
    Analogous to the builtin :func:`iter()`.
    """

    if sentinel is _sentinel:
        # Since we cannot directly return the return value of obj.__aiter__()
        # as being an async-generator, we do the async-iteration here.
        async for item in obj:
            yield item
    else:
        while True:
            item = await obj()
            if item == sentinel:
                break
            yield item



"""
Provides a Golang-like ``defer()`` API using decorators, which allows grouping
resource initialization and cleanup in one place without extra indentations.

Example:

.. code-block:: python3

   async def init(x):
       ...

   async def cleanup(x):
       ...

   @aiotools.adefer
   async def do(defer):  # <-- be aware of defer argument!
       x = SomeResource()
       await init(x)
       defer(cleanup(x))
       ...
       ...

This is equivalent to:

.. code-block:: python3

   async def do():
       x = SomeResource()
       await init(x)
       try:
           ...
           ...
       finally:
           await cleanup(x)

Note that :class:`aiotools.context.AsyncContextGroup` or
:class:`contextlib.AsyncExitStack` serves well for the same purpose, but for simple
cleanups, this defer API makes your codes simple because it steps aside the main
execution context without extra indentations.

.. warning::

   Any exception in the deferred functions is raised transparently, and may block
   execution of the remaining deferred functions.
   This behavior may be changed in the future versions, though.
"""

from collections import deque
import functools
import inspect
from typing import (
    Union,
    Awaitable,
    Callable,
)

__all__ = (
    'defer', 'adefer',
)


def defer(func):
    """
    A synchronous version of the defer API.
    It can only defer normal functions.
    """
    assert not inspect.iscoroutinefunction(func), \
           'the decorated function must not be async'

    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        deferreds = deque()

        def defer(f: Callable) -> None:
            assert not inspect.iscoroutinefunction(f), \
                   'the deferred function must not be async'
            assert not inspect.iscoroutine(f), \
                   'the deferred object must not be a coroutine'
            deferreds.append(f)

        try:
            return func(defer, *args, **kwargs)
        finally:
            while deferreds:
                f = deferreds.pop()
                f()

    return _wrapped


def adefer(func):
    """
    An asynchronous version of the defer API.
    It can defer coroutine functions, coroutines, and normal functions.
    """
    assert inspect.iscoroutinefunction(func), \
           'the decorated function must be async'

    @functools.wraps(func)
    async def _wrapped(*args, **kwargs):
        deferreds = deque()

        def defer(f: Union[Callable, Awaitable]) -> None:
            deferreds.append(f)

        try:
            return await func(defer, *args, **kwargs)
        finally:
            while deferreds:
                f = deferreds.pop()
                if inspect.iscoroutinefunction(f):
                    await f()
                elif inspect.iscoroutine(f):
                    await f
                else:
                    f()

    return _wrapped



"""
Provides an implementation of asynchronous context manager and its applications.

.. note::

   The async context managers in this module are transparent aliases to
   ``contextlib.asynccontextmanager`` of the standard library in Python 3.7
   and later.
"""

import abc
import contextlib
import asyncio
import functools
import inspect
from typing import Any, Callable, Iterable, Optional, List

__all__ = [
    'AsyncContextManager', 'async_ctx_manager', 'actxmgr', 'aclosing',
    'AsyncContextGroup', 'actxgroup',
]


if hasattr(contextlib, 'asynccontextmanager'):
    __all__ += ['AsyncExitStack']

    AbstractAsyncContextManager = \
        contextlib.AbstractAsyncContextManager
    AsyncContextManager = \
        contextlib._AsyncGeneratorContextManager     # type: ignore
    AsyncExitStack = contextlib.AsyncExitStack
    async_ctx_manager = contextlib.asynccontextmanager
else:
    __all__ += ['AsyncContextDecorator', 'actxdecorator']

    class AbstractAsyncContextManager(abc.ABC):  # type: ignore
        """
        The base abstract interface for asynchronous context manager.
        """

        async def __aenter__(self):
            return self  # pragma: no cover

        @abc.abstractmethod
        async def __aexit__(self, exc_type, exc_value, tb):
            return None  # pragma: no cover

        @classmethod
        def __subclasshook__(cls, C):
            if cls is AbstractAsyncContextManager:
                if (any('__aenter__' in B.__dict__ for B in C.__mro__) and
                    any('__aexit__' in B.__dict__ for B in C.__mro__)):
                    return True
            return NotImplemented

    class AsyncContextDecorator:
        """
        Make an asynchronous context manager be used as a decorator function.
        """

        def _recreate_cm(self):
            return self

        def __call__(self, func):
            @functools.wraps(func)
            async def inner(*args, **kwargs):
                async with self._recreate_cm():
                    return (await func(*args, **kwargs))
            return inner

    actxdecorator = AsyncContextDecorator

    class AsyncContextManager(AsyncContextDecorator,         # type: ignore
                              AbstractAsyncContextManager):
        """
        Converts an async-generator function into asynchronous context manager.
        """

        def __init__(self, func: Callable[..., Any], args, kwargs):
            if not inspect.isasyncgenfunction(func):
                raise RuntimeError('Context manager function must be '
                                   'an async-generator')
            self._agen = func(*args, **kwargs)
            self.func = func
            self.args = args
            self.kwargs = kwargs

        def _recreate_cm(self):
            return self.__class__(self.func, self.args, self.kwargs)

        async def __aenter__(self):
            try:
                return (await self._agen.__anext__())
            except StopAsyncIteration:
                # The generator should yield at least once.
                raise RuntimeError("async-generator didn't yield") from None

        async def __aexit__(self, exc_type, exc_value, tb):
            if exc_type is None:
                # This is the normal path when the context body
                # did not raise any exception.
                try:
                    await self._agen.__anext__()
                except StopAsyncIteration:
                    return
                else:
                    # The generator has already yielded,
                    # no more yields are allowed.
                    raise RuntimeError("async-generator didn't stop") from None
            else:
                # The context body has raised an exception.
                if exc_value is None:
                    # Ensure exc_value is a valid Exception.
                    exc_value = exc_type()
                try:
                    # Throw the catched exception into the generator,
                    # so that it can handle as it wants.
                    await self._agen.athrow(exc_type, exc_value, tb)
                    # Here the generator should have finished!
                    # (i.e., it should not yield again in except/finally blocks!)
                    raise RuntimeError("async-generator didn't stop after athrow()")
                    # NOTE for PEP-479
                    #   StopAsyncIteration raised inside the context body
                    #   is converted to RuntimeError.
                    #   In the standard library's contextlib.py, there is
                    #   an extra except clause to catch StopIteration here,
                    #   but this is unnecessary now.
                except StopAsyncIteration as exc_new_value:
                    return exc_new_value is not exc_value
                except RuntimeError as exc_new_value:
                    # When the context body did not catch the exception, re-raise.
                    if exc_new_value is exc_value:
                        return False
                    # When the context body's exception handler raises
                    # another chained exception, re-raise.
                    if isinstance(exc_value, (StopIteration, StopAsyncIteration)):
                        if exc_new_value.__cause__ is exc_value:
                            return False
                    # If this is a purely new exception, raise the new one.
                    raise
                except (BaseException, asyncio.CancelledError) as exc:
                    if exc is not exc_value:
                        raise

    def async_ctx_manager(func):
        @functools.wraps(func)
        def helper(*args, **kwargs):
            return AsyncContextManager(func, args, kwargs)
        return helper


class aclosing:
    """
    An analogy to :func:`contextlib.closing` for async generators.

    The motivation has been proposed by:

    * https://github.com/njsmith/async_generator
    * https://vorpus.org/blog/some-thoughts-on-asynchronous-api-design-\
in-a-post-asyncawait-world/#cleanup-in-generators-and-async-generators
    * https://www.python.org/dev/peps/pep-0533/
    """

    def __init__(self, thing):
        self.thing = thing

    async def __aenter__(self):
        return self.thing

    async def __aexit__(self, *args):
        await self.thing.aclose()


class AsyncContextGroup:
    """
    Merges a group of context managers into a single context manager.
    Internally it uses :func:`asyncio.gather()` to execute them with overlapping,
    to reduce the execution time via asynchrony.

    Upon entering, you can get values produced by the entering steps from
    the passed context managers (those ``yield``-ed) using an ``as`` clause of
    the ``async with``
    statement.

    After exits, you can check if the context managers have finished
    successfully by ensuring that the return values of ``exit_states()`` method
    are ``None``.

    .. note::

       You cannot return values in context managers because they are
       generators.

    If an exception is raised before the ``yield`` statement of an async
    context manager, it is stored at the corresponding manager index in the
    as-clause variable.  Similarly, if an exception is raised after the
    ``yield`` statement of an async context manager, it is stored at the
    corresponding manager index in the ``exit_states()`` return value.

    Any exceptions in a specific context manager does not interrupt others;
    this semantic is same to ``asyncio.gather()``'s when
    ``return_exceptions=True``.  This means that, it is user's responsibility
    to check if the returned context values are exceptions or the intended ones
    inside the context body after entering.

    :param context_managers: An iterable of async context managers.
                             If this is ``None``, you may add async context
                             managers one by one using the :meth:`~.add`
                             method.

    """

    def __init__(self,
                 context_managers: Optional[Iterable[AbstractAsyncContextManager]] = None):  # noqa
        self._cm = list(context_managers) if context_managers else []
        self._cm_yields: List[asyncio.Task] = []
        self._cm_exits: List[asyncio.Task] = []

    def add(self, cm):
        """
        TODO: fill description
        """
        self._cm.append(cm)

    async def __aenter__(self):
        # Exceptions in context managers are stored into _cm_yields list.
        # NOTE: There is no way to "skip" the context body even if the entering
        #       process fails.
        self._cm_yields[:] = await asyncio.gather(
            *(e.__aenter__() for e in self._cm),
            return_exceptions=True)
        return self._cm_yields

    async def __aexit__(self, *exc_info):
        # Clear references to context variables.
        self._cm_yields.clear()
        # Exceptions are stored into _cm_exits list.
        self._cm_exits[:] = await asyncio.gather(
            *(e.__aexit__(*exc_info) for e in self._cm),
            return_exceptions=True)

    def exit_states(self):
        """
        TODO: fill description
        """
        return self._cm_exits


# Shorter aliases
actxmgr = async_ctx_manager
actxgroup = AsyncContextGroup



import collections
import functools

from .compat import get_running_loop

__all__ = (
    'apartial',
    'lru_cache',
)

_CacheEntry = collections.namedtuple('_CacheEntry', 'value expire_at')


def apartial(coro, *args, **kwargs):
    """
    Wraps a coroutine function with pre-defined arguments (including keyword
    arguments).  It is an asynchronous version of :func:`functools.partial`.
    """

    @functools.wraps(coro)
    async def wrapped(*cargs, **ckwargs):
        return await coro(*args, *cargs, **kwargs, **ckwargs)

    return wrapped


def lru_cache(maxsize: int = 128,
              typed: bool = False,
              expire_after: float = None):
    """
    A simple LRU cache just like :func:`functools.lru_cache`, but it works for
    coroutines.  This is not as heavily optimized as :func:`functools.lru_cache`
    which uses an internal C implementation, as it targets async operations
    that take a long time.

    It follows the same API that the standard functools provides.  The wrapped
    function has ``cache_clear()`` method to flush the cache manually, but
    leaves ``cache_info()`` for statistics unimplemented.

    Note that calling the coroutine multiple times with the same arguments
    before the first call returns may incur duplicate executions.

    This function is not thread-safe.

    Args:
       maxsize: The maximum number of cached entries.

       typed: Cache keys in different types separately (e.g., ``3`` and ``3.0`` will
              be different keys).

       expire_after: Re-calculate the value if the configured time has passed even
                     when the cache is hit.  When re-calculation happens the
                     expiration timer is also reset.
    """

    if maxsize is not None and not isinstance(maxsize, int):
        raise TypeError('Expected maxsize to be an integer or None')

    def wrapper(coro):

        sentinel = object()  # unique object to distinguish None as result
        cache = collections.OrderedDict()
        cache_get = cache.get
        cache_del = cache.__delitem__
        cache_set = cache.__setitem__
        cache_len = cache.__len__
        cache_move = cache.move_to_end
        make_key = functools._make_key

        # We don't use explicit locks like the standard functools,
        # because this lru_cache is intended for use in asyncio coroutines.
        # The only context interleaving happens when calling the user-defined
        # coroutine, so there is no need to add extra synchronization guards.

        @functools.wraps(coro)
        async def wrapped(*args, **kwargs):
            now = get_running_loop().time()
            k = make_key(args, kwargs, typed)
            entry = cache_get(k, sentinel)
            if entry is not sentinel:
                if entry.expire_at is None:
                    return entry.value
                if entry.expire_at >= now:
                    return entry.value
                cache_del(k)
            result = await coro(*args, **kwargs)
            if maxsize is not None and cache_len() >= maxsize:
                cache.popitem(last=False)
            if expire_after is not None:
                expire_at = now + expire_after
            else:
                expire_at = None
            cache_set(k, _CacheEntry(result, expire_at))
            cache_move(k, last=True)
            return result

        def cache_clear():
            cache.clear()

        def cache_info():
            raise NotImplementedError

        wrapped.cache_clear = cache_clear
        wrapped.cache_info = cache_info
        return wrapped

    return wrapper
